use crate::store::{OwnerSource, Visibility};
use chrono::{DateTime, Utc};
use url::Url;
use uuid::Uuid;

mod apub;
mod block;
mod comment;
mod follow;
mod follow_request;
mod profile;
mod react;
mod submission;

pub(crate) use apub::ingest;

pub struct CreateReact {
    like_apub_id: Option<Url>,
    submission_id: Uuid,
    profile_id: Uuid,
    comment_id: Option<Uuid>,
    react: String,
    published: DateTime<Utc>,
}

pub struct DeleteReact {
    react_id: Uuid,
}

pub struct CreateComment {
    note_apub_id: Option<Url>,
    submission_id: Uuid,
    profile_id: Uuid,
    comment_id: Option<Uuid>,
    body: String,
    published: DateTime<Utc>,
}

pub struct AnnounceComment;

pub struct UpdateComment {
    update_apub_id: Option<Url>,
    comment_id: Uuid,
    body: Option<String>,
}

pub struct DeleteComment {
    note_apub_id: Url,
    comment_id: Uuid,
}

pub struct CreateSubmission {
    note_apub_id: Option<Url>,
    profile_id: Uuid,
    title: String,
    description: Option<String>,
    files: Vec<Uuid>,
    published: DateTime<Utc>,
    visibility: Visibility,
}

pub struct AnnounceSubmission;

pub struct UpdateSubmission {
    submission_id: Uuid,
    title: Option<String>,
    description: Option<String>,
    files: Vec<Uuid>,
}

pub struct DeleteSubmission {
    submission_id: Uuid,
}

pub struct CreateProfileApub {
    person_apub_id: Url,
    public_key_id: Url,
    public_key: String,
}

pub struct CreateProfile {
    apub: Option<CreateProfileApub>,
    owner_source: OwnerSource,
    handle: String,
    domain: String,
    display_name: Option<String>,
    description: Option<String>,
    login_required: bool,
    icon: Option<Uuid>,
    banner: Option<Uuid>,
    published: DateTime<Utc>,
}

impl CreateProfile {
    pub fn from_local(owner_id: Uuid, handle: String, domain: String) -> Self {
        CreateProfile {
            apub: None,
            owner_source: OwnerSource::Local(owner_id),
            handle,
            domain,
            display_name: None,
            description: None,
            login_required: true,
            icon: None,
            banner: None,
            published: Utc::now(),
        }
    }
}

pub struct UpdateProfile {
    profile_id: Uuid,
    display_name: Option<String>,
    description: Option<String>,
    login_required: Option<bool>,
    icon: Option<Uuid>,
    banner: Option<Uuid>,
    public_key_id: Option<Url>,
    public_key: Option<String>,
}

impl UpdateProfile {
    pub fn from_text(profile_id: Uuid, display_name: String, description: String) -> Self {
        UpdateProfile {
            profile_id,
            display_name: Some(display_name),
            description: Some(description),
            login_required: None,
            icon: None,
            banner: None,
            public_key_id: None,
            public_key: None,
        }
    }

    pub fn from_icon(profile_id: Uuid, icon: Uuid) -> Self {
        UpdateProfile {
            profile_id,
            display_name: None,
            description: None,
            login_required: None,
            icon: Some(icon),
            banner: None,
            public_key_id: None,
            public_key: None,
        }
    }

    pub fn from_banner(profile_id: Uuid, banner: Uuid) -> Self {
        UpdateProfile {
            profile_id,
            display_name: None,
            description: None,
            login_required: None,
            icon: None,
            banner: Some(banner),
            public_key_id: None,
            public_key: None,
        }
    }

    pub fn from_login_required(profile_id: Uuid, login_required: bool) -> Self {
        UpdateProfile {
            profile_id,
            display_name: None,
            description: None,
            login_required: Some(login_required),
            icon: None,
            banner: None,
            public_key_id: None,
            public_key: None,
        }
    }
}

pub struct DeleteProfile {
    profile_id: Uuid,
}

pub struct CreateFollowRequest {
    follow_apub_id: Option<Url>,
    followed_profile: Uuid,
    followed_by_profile: Uuid,
    published: DateTime<Utc>,
}

pub struct AcceptFollowRequest {
    accept_apub_id: Option<Url>,
    follow_request_id: Uuid,
    published: DateTime<Utc>,
}

pub struct RejectFollowRequest {
    follow_request_id: Uuid,
}

pub struct UndoFollowRequest {
    follow_request_id: Uuid,
}

pub struct CreateBlock {
    block_apub_id: Option<Url>,
    blocked_profile: Uuid,
    blocked_by_profile: Uuid,
    published: DateTime<Utc>,
}

pub struct DeleteBlock {
    block_id: Uuid,
}

pub struct UndoFollow {
    follow_apub_id: Url,
    follow_id: Uuid,
}

pub struct UndoAcceptFollow {
    follow_apub_id: Url,
    follow_id: Uuid,
}
